const unique ArrayList.footprint: <[ref]bool>name;

procedure ArrayList..ctor(this: ref);
  free requires IsGoodObject(this,Heap);
  modifies Heap;
  // user-declared postconditions
  ensures #ArrayList.size(Heap,this) == 0;
  // fresh(this.footprint - { this });
  ensures (forall p: ref:: Heap[this,ArrayList.footprint][p] 
             ==> (p == this || !old(Heap)[p,alloc]));

  // target object is allocated upon return
  free ensures IsGoodObject(this, Heap);
  // frame condition
  ensures 
      (forall p: ref, f: name ::
        old(Heap)[p, alloc] ==> old(Heap[p,f]) == Heap[p,f] || p == this 
      );
  ensures ArrayList.INV(Heap,this);
  free ensures NoDeallocs(old(Heap),Heap);

function ArrayList.INV(h: [ref, <x>name]x, o: ref) returns (bool);

function #ArrayList.size(h: [ref, <x>name]x, o: ref) returns (int);

procedure ArrayList.Add(this: ref, o: any);
  free requires IsGoodObject(this,Heap);
  requires ArrayList.INV(Heap,this);
  modifies Heap;
  ensures ArrayList.INV(Heap,this);
  ensures #ArrayList.size(Heap,this) == #ArrayList.size(old(Heap),this) + 1;
  ensures #ArrayList.get(Heap,this,#ArrayList.size(Heap,this) - 1) == o;
  ensures (forall i:int :: 
             0 <= i && i < #ArrayList.size(Heap,this) - 1
               ==>
             #ArrayList.get(Heap,this,i) == old(#ArrayList.get(Heap,this,i)));
             
  // frame condition: "modifies footprint"
  ensures 
      (forall p: ref, f: name ::
        old(Heap)[p, alloc] ==> 
         old(Heap[p,f]) == Heap[p,f] // everything remains unchanged
         || old(Heap)[this,ArrayList.footprint][p] // unless it is 
         //an object contained in self's footprint
      );
  ensures DifferenceIsNew(old(Heap), 
                          old(Heap)[this,ArrayList.footprint], 
                          Heap[this, ArrayList.footprint]);
  free ensures NoDeallocs(old(Heap),Heap);
  free ensures IsGoodHeap(Heap);

function #ArrayList.get(h: [ref, <x>name]x, o: ref, i : int) returns (any);

procedure ArrayList.Get(this: ref, index: int) returns (result: any);
  requires ArrayList.INV(Heap,this);
  requires 0 <= index && index < #ArrayList.size(Heap,this);
  ensures result == #ArrayList.get(Heap,this,index);

