/* Verification of Client using SubX and SubObsX (Section 6.1 of paper) */
/* Hidden module invariant is not used */

var $Heap: [ref, <x>name]x where IsGoodHeap($Heap);

// These are the fields of the class Subject
const unique Subject.obs: <ref>name;
const unique Subject.val: <int>name;
const unique Subject.O: <[ref]bool>name;

// These are the fields of the class Observer
const unique Observer.sub: <ref>name;
const unique Observer.cache: <int>name;
const unique Observer.nxt: <ref>name;

function SubX(s : ref, v : int, h: [ref,<x>name]x) returns (bool);
axiom (forall s : ref, v : int, h: [ref,<x>name]x :: { SubX(s,v,h) }
        SubX(s,v,h)  <==> (IsGoodObject(s,h) && h[s,Subject.val] == v)
      );
      
function Obs(o : ref, s : ref, v : int, h: [ref,<x>name]x) returns (bool);
axiom (forall o : ref, s : ref, v : int, h: [ref,<x>name]x :: { Obs(o,s,v,h) }
        Obs(o,s,v,h)  <==> IsGoodObject(o,h) && IsGoodObject(s,h) && 
                           h[o,Observer.cache] == v && h[o,Observer.sub] == s
      );
      
function SubObs(s : ref, v : int, h: [ref,<x>name]x) returns (bool);
axiom (forall s : ref, v : int, h: [ref,<x>name]x :: { SubObs(s,v,h) }
        SubObs(s,v,h) <==> Sub(s,v,h) && 
                           (forall o : ref :: IsGoodObject(o,h) && h[s,Subject.O][o] 
                                              ==> Obs(o,s,v,h))
      );

axiom (forall h: [ref,<x>name]x, s: ref :: { IsGoodRegion(h[s, Subject.O], h) }
               IsGoodObject(s, h) ==> IsGoodRegion(h[s, Subject.O], h)
      );

axiom (forall h: [ref,<x>name]x, s: ref, o: ref :: { h[s, Subject.O][o] }
               IsGoodObject(s, h) ==> IsGoodRegion(h[s, Subject.O], h)
      );

procedure Subject..ctor(this: ref);
    free requires IsGoodObject(this,$Heap);
    modifies $Heap;
    // user-declared postconditions
    ensures SubObsX(this,0,$Heap);
    ensures EqualSet($Heap[this, Subject.O], EmptySet);
    // target object is allocated upon return
    free ensures $Heap[this, alloc];
    // wr <this>\Dot val, <this>\Dot obs, <this>\Dot O

    ensures 
      (forall p: ref, f: name ::
        IsGoodObject(p,old($Heap)) ==>
        old($Heap[p,f]) == $Heap[p,f] || p == this 
      );
    free ensures NoDeallocs(old($Heap), $Heap);

/* Subject.register and Subject.add are MODULE-SCOPED hence omitted*/

procedure Subject.update(this: ref, n: int);
    // user-declared preconditions
    free requires IsGoodObject(this, $Heap);
    requires SubObsX(this,$Heap[this,Subject.val],$Heap);
    modifies $Heap;
    // user-declared postconditions
    ensures SubObsX(this, n, $Heap);
    // wr <this>\Dot val, <this>.O\Dot cache

    ensures
      (forall p: ref, f: name ::
        IsGoodObject(p,old($Heap)) ==>
        old($Heap[p,f]) == $Heap[p,f] || 
          (p == this && f == Subject.val) ||
          ($Heap[this,Subject.O][p] && f == Observer.cache)
      );

    free ensures NoDeallocs(old($Heap), $Heap);


procedure Subject.get(this: ref) returns ($result : int);
    free requires IsGoodObject(this, $Heap);
    // user-declared preconditions
    requires SubX(this,$Heap[this,Subject.val],$Heap);
    // user-declared postconditions
    ensures SubX(this,$Heap[this,Subject.val],$Heap);
    ensures $result == $Heap[this,Subject.val];

procedure Observer..ctor(this: ref, s : ref);
    free requires IsGoodObject(s, $Heap);
    // needed because we are passing self to register 
    free requires IsGoodObject(this,$Heap); 
    requires (forall p : ref :: IsGoodObject(p, $Heap) ==> !$Heap[p, Subject.O][this]);

    requires SubX(s, $Heap[s,Subject.val], $Heap);
    modifies $Heap;
    // user-declared postconditions
    ensures SubX(s, $Heap[s,Subject.val], $Heap);
    ensures Obs(this, s, $Heap[s,Subject.val], $Heap);

    ensures $Heap[s,Subject.O][this];
    //needed for client
    ensures EqualSet(UnionSet(SingletonSet(this), old($Heap)[s,Subject.O]), 
                     $Heap[s,Subject.O]) ; 
    // wr s.O\Dot nxt, <s>\Dot O
    // Since the only nxt field that is changed is "this" one, don't need to own
    // up to it here in the modifies clause.
    ensures 
      (forall p: ref, f: name ::
        IsGoodObject(p,old($Heap)) ==>
        old($Heap[p,f]) == $Heap[p,f] || p == this ||
          (old($Heap)[s,Subject.O][p] && f == Observer.nxt) || 
          (p == s && f == Subject.O) ||
          (p == s && f == Subject.obs)          /* for abstraction, Subject.any */
      );
    free ensures NoDeallocs(old($Heap), $Heap);


procedure Observer.notify(this: ref);
    free requires IsGoodObject(this,$Heap);
    // user-declared preconditions
    requires SubX($Heap[this,Observer.sub],
                  $Heap[$Heap[this,Observer.sub],Subject.val],
                  $Heap);
    modifies $Heap;
    // user-declared postconditions
    ensures SubX($Heap[this,Observer.sub],
                 $Heap[$Heap[this,Observer.sub],Subject.val],
                 $Heap);
    ensures Obs(this,$Heap[this,Observer.sub],
                $Heap[$Heap[this,Observer.sub],Subject.val], $Heap);
    // wr <this>.\Dot cache

    ensures 
      (forall p: ref, f: name :: { $Heap[p,f] }
        IsGoodObject(p,old($Heap)) ==>
        old($Heap[p,f]) == $Heap[p,f] || 
           (p == this && f == Observer.cache) 
      );

    free ensures NoDeallocs(old($Heap), $Heap);


procedure Observer.val(this: ref) returns ($result : int);
    free requires IsGoodObject(this,$Heap);
    // user-declared preconditions
    requires SubObsX($Heap[this,Observer.sub], 
                     $Heap[$Heap[this,Observer.sub],Subject.val],$Heap);
    requires $Heap[$Heap[this,Observer.sub],Subject.O][this];
    // user-declared postconditions
    ensures SubObsX($Heap[this,Observer.sub], 
                    $Heap[$Heap[this,Observer.sub],Subject.val],
                    $Heap);
    ensures $result == $Heap[$Heap[this,Observer.sub],Subject.val];

////////////////////////////////////////////////////////////////////
//////////////     /////CLIENT//////////////////////////////////////
////////////////////////////////////////////////////////////////////
procedure client()
    modifies $Heap;
    // user-declared postconditions
    // effects
  {   
        var sub0 : ref;
        var obs0 : ref;
        var obs1 : ref;
        var sub  : ref;
        var obs  : ref;

    // sub0 = new Subject();
        havoc sub0;
        assume !$Heap[sub0,alloc];
        call Subject..ctor(sub0);

    // obs0 = new Observer(sub0);
        havoc obs0;
        assume !$Heap[obs0,alloc];
        assume EqualSet($Heap[obs0, Subject.O],  EmptySet);
        call Observer..ctor(obs0, sub0);

    // obs1 = new Observer(sub0);
        havoc obs1;
        assume !$Heap[obs1,alloc];
        assume EqualSet($Heap[obs1, Subject.O],  EmptySet);
        call Observer..ctor(obs1, sub0);

   //  sub  = new Subject();
        havoc sub;
        assume !$Heap[sub,alloc];
        call Subject..ctor(sub);

   //  obs = new Observer(sub);
        havoc obs;
        assume !$Heap[obs,alloc];
        assume EqualSet($Heap[obs, Subject.O],  EmptySet);
        call Observer..ctor(obs, sub);

        call Subject.update(sub0, 5);
        assert DisjointSet($Heap[sub0, Subject.O], $Heap[sub, Subject.O]);
        assert SubObsX(sub0, 5, $Heap);
        assert SubObsX(sub, 0, $Heap);
  }
