// List Copy
var $Heap: [ref, <x>name]x where IsGoodHeap($Heap);

// This is the "nxt" field of the class Node, used for linked lists.
const unique Node.nxt: <ref>name;

procedure Node..ctor(this: ref );
  free requires IsGoodHeap($Heap);
  modifies $Heap;
  // user-declared postconditions
  ensures $Heap[this, Node.nxt] == null;
  // target object is allocated upon return
  free ensures IsGoodObject(this, $Heap);
  ensures 
      (forall p: ref, f: name ::
        old($Heap)[p, alloc] ==> old($Heap[p,f]) == $Heap[p,f] || p == this 
      );
  free ensures NoDeallocs(old($Heap),$Heap);
  
  // invariant: r.nxt < r;
  //         forall Node n in r : n.nxt < r
  // invariant head == null || head < r

procedure ListCopy(root: ref) returns ($result: ref)
    free requires IsGoodHeap($Heap);
    free requires (root == null) || IsGoodObject(root, $Heap);

    modifies $Heap;
    
    free ensures IsGoodHeap($Heap);
    free ensures NoDeallocs(old($Heap),$Heap);
  {
    var newRoot: ref, prev: ref, tmp: ref, oldListPtr: ref; 
    var newRegion: [ref]bool where IsGoodRegion(newRegion,$Heap);
    var existingRegion: [ref]bool where IsGoodRegion(existingRegion, $Heap);
    var h: [ref,<x>name]x;
    var loopPreHeaderHeap: [ref,<x>name]x;

    // simulate preconditions for things not visible outside of the method
    assume root == null || existingRegion[root]; 
    assume RegionClosure(existingRegion,$Heap,Node.nxt);
    newRoot := null;
    tmp := null;
    oldListPtr := root;
    newRegion := EmptySet;
    if (oldListPtr != null) {
      // newRoot = new Node();
        havoc newRoot;
        assume !$Heap[newRoot,alloc];
        call Node..ctor(newRoot); // after this call IsGoodObject(newRoot) holds
      newRegion := UnionSet(newRegion, SingletonSet(newRoot)); //r = r+<newRoot>;
      prev := newRoot;
      loopPreHeaderHeap := $Heap;
      oldListPtr := $Heap[oldListPtr,Node.nxt]; //root = root.nxt;
      while (oldListPtr != null)
        invariant newRegion[newRoot];
        invariant RegionClosure(newRegion, $Heap, Node.nxt);
        invariant fresh(old($Heap), $Heap, newRegion);
        invariant NoDeallocs(loopPreHeaderHeap, $Heap); 
      {
        // tmp = new Node();
          havoc tmp;
          assume !$Heap[tmp,alloc];
          call Node..ctor(tmp); // after this call IsGoodObject(tmp) holds
        //r = r+<tmp>;
        newRegion := UnionSet(newRegion,SingletonSet(tmp)); 
        // prev.nxt = tmp; tmp goes on the end of the growing new list
        $Heap[prev,Node.nxt] := tmp; 
        assume IsGoodHeap($Heap); // need after every heap update
        prev := tmp;
        //root = root.nxt;
        oldListPtr := $Heap[oldListPtr,Node.nxt]; 
      }
    } 
    $result := newRoot;
    
    // Simulated post-conditions because "newRegion" is not visible in the method contract
    assert $result == null || newRegion[$result];
    assert RegionClosure(newRegion,$Heap,Node.nxt);
    // everything in newRegion is fresh
    assert fresh(old($Heap), $Heap, newRegion);
    // newRegion # exisitingRegion
    assert DisjointSet(newRegion, existingRegion); 
    assert IsGoodRegion(newRegion, $Heap);
    
  }
  
