// Multiple Subject Observer: an observer can observe multiple subjects
var $Heap: [ref, <x>name]x where IsGoodHeap($Heap);

// These are the fields of the class Node
const unique Node.info: <ref>name; // reference to an Observer
const unique Node.nxt: <ref>name;

// These are the fields of the class Subject
const unique Subject.obs: <ref>name; // pointer to the first Node 
      //holding the list of observers
const unique Subject.val: <int>name;
const unique Subject.O: <[ref]bool>name;

// These are the fields of the class Observer
// Observer.caches is now an array 
// maps a subject to the cached value for that subject
const unique Observer.caches: <[ref]int>name; 

// In Spec#, these would be from the translation of [Pure] methods 
// written as part of the program
function List2Observers(o : ref, h: [ref,<x>name]x) returns ([ref]bool);
axiom (forall o : ref, h: [ref,<x>name]x :: { List2Observers(o,h) }
        (o == null <==> EqualSet(List2Observers(o,h),EmptySet))
      );

axiom (forall o : ref, h: [ref,<x>name]x, r: [ref]bool :: 
        { EqualSet(List2Observers(o,h), r) } 
        (o != null ==> EqualSet(List2Observers(o,h), 
                                UnionSet(SingletonSet(h[o,Node.info]),
                                         List2Observers(h[o,Node.nxt],h))))
      );

axiom (forall H: [ref,<x>name]x, K: [ref,<x>name]x, o : ref ::
         { List2Observers(o,H), List2Observers(o,K) }
         (forall o2 : ref  ::
            o2 != null && List2Observers(o,H)[o2] && List2Observers(o,K)[o2] 
            ==> H[o2,Observer.nxt] == K[o2,Observer.nxt])
          ==>
          EqualSet(List2Observers(o,H), List2Observers(o,K))
      );

function Sub(s : ref, v : int, h: [ref,<x>name]x) returns (bool);
axiom (forall s : ref, v : int, h: [ref,<x>name]x :: { Sub(s,v,h) }
        Sub(s,v,h)  <==> (IsGoodObject(s,h) && h[s,Subject.val] == v && 
                          EqualSet(List2Observers(h[s,Subject.obs],h),
                                   h[s,Subject.O]))
      );
      
function Obs(o : ref, s : ref, v : int, h: [ref,<x>name]x) returns (bool);
axiom (forall o : ref, s : ref, v : int, h: [ref,<x>name]x :: { Obs(o,s,v,h) }
        Obs(o,s,v,h)  <==> IsGoodObject(o,h) && IsGoodObject(s,h) && 
                           h[o,Observer.caches][s] == v
      );
      
function SubObs(s : ref, v : int, h: [ref,<x>name]x) returns (bool);
axiom (forall s : ref, v : int, h: [ref,<x>name]x :: { SubObs(s,v,h) }
        SubObs(s,v,h) <==> Sub(s,v,h) && 
                           (forall o : ref :: IsGoodObject(o,h) && h[s,Subject.O][o] 
                            ==> Obs(o,s,v,h))
      );

axiom (forall h: [ref,<x>name]x, s: ref :: { IsGoodRegion(h[s, Subject.O], h) }
               IsGoodObject(s, h) ==> IsGoodRegion(h[s, Subject.O], h)
      );

axiom (forall h: [ref,<x>name]x, s: ref, o: ref :: { h[s, Subject.O][o] }
               IsGoodObject(s, h) ==> IsGoodRegion(h[s, Subject.O], h)
      );

// NODE
procedure Node..ctor(this: ref, element: ref where IsGoodObject(element,$Heap));
  free requires IsGoodHeap($Heap);
  modifies $Heap;
  // user-declared postconditions
  ensures $Heap[this, Node.nxt] == null;
  ensures $Heap[this, Node.info] == element;
  // target object is allocated upon return
  free ensures IsGoodObject(this, $Heap);
  ensures 
      (forall p: ref, f: name ::
        old($Heap)[p, alloc] ==> old($Heap[p,f]) == $Heap[p,f] || p == this 
      );
  free ensures NoDeallocs(old($Heap),$Heap);
// END OF NODE

// SUBJECT
procedure Subject..ctor(this: ref)
    free requires IsGoodObject(this,$Heap);
    modifies $Heap;
    // user-declared postconditions
    ensures SubObs(this,0,$Heap);
    ensures EqualSet($Heap[this, Subject.O], EmptySet);
    // target object is allocated upon return
    free ensures $Heap[this, alloc];
    // wr <this>\Dot val, <this>\Dot obs, <this>\Dot O
    ensures 
      (forall p: ref, f: name ::
        IsGoodObject(p,old($Heap)) ==>
        old($Heap[p,f]) == $Heap[p,f] || p == this 
      );
    free ensures NoDeallocs(old($Heap), $Heap);

  {
    $Heap[this,Subject.obs] := null;
    assume IsGoodHeap($Heap);  
    $Heap[this,Subject.val] := 0;
    assume IsGoodHeap($Heap);
    $Heap[this,Subject.O] := EmptySet;
    assume IsGoodHeap($Heap);
  }

procedure Subject.register(this: ref where IsGoodObject(this, $Heap), 
                           o: ref where IsGoodObject(o, $Heap))
    // user-declared preconditions
    // to prevent o from being added to the list twice */
    requires !$Heap[this, Subject.O][o]; 
    requires Sub(this, $Heap[this,Subject.val], $Heap);
    modifies $Heap;

    // user-declared postconditions

    ensures Sub(this, $Heap[this,Subject.val], $Heap);
    ensures Obs(o,this,$Heap[this,Subject.val],$Heap);
    ensures $Heap[this,Subject.O][o];
    ensures EqualSet(UnionSet(SingletonSet(o),old($Heap)[this,Subject.O]),
                     $Heap[this,Subject.O]) ;

    // wr <this>\Dot O, (<o>\Dot caches)[this], <this>\Dot obs, wr all
    ensures 
      (forall p: ref, f: name ::
        IsGoodObject(p,old($Heap)) ==>
        old($Heap[p,f]) == $Heap[p,f] || 
          (p == this && f == Subject.O) ||
          (p == o && f == Observer.caches) ||
          (p == this && f == Subject.obs)
      );
    // Part of "wr (<o>\Dot caches)[this]": 
    // no other cache line has changed except for "this" one.
    ensures (forall s : ref ::
              IsGoodObject(s,$Heap) && s != this
                ==>
              $Heap[o,Observer.caches][s] == old($Heap)[o,Observer.caches][s]
            );
    free ensures NoDeallocs(old($Heap), $Heap);
  {
    call Subject.add(this,o);
    call Observer.notify(o, this);
  }

procedure Subject.update(this: ref, n: int)
    // user-declared preconditions
    free requires IsGoodObject(this, $Heap);
    requires SubObs(this,$Heap[this,Subject.val],$Heap);
    modifies $Heap;

    // user-declared postconditions
    ensures SubObs(this, n, $Heap);
    // wr <this>\Dot val, (<this>.O\Dot caches)[this] 
    ensures
      (forall p: ref, f: name ::
        IsGoodObject(p,old($Heap)) ==>
        old($Heap[p,f]) == $Heap[p,f] || 
          (p == this && f == Subject.val) ||
          (old($Heap)[this,Subject.O][p] && f == Observer.caches)
      );
    free ensures NoDeallocs(old($Heap), $Heap);
  {
    var node : ref where (node == null) || IsGoodObject(node, $Heap); 
    var r1 : [ref]bool where IsGoodRegion(r1, $Heap);

    $Heap[this,Subject.val] := n;
    assume IsGoodHeap($Heap);
    node := $Heap[this,Subject.obs];
    r1 := EmptySet;
    while (node != null)
      invariant Sub(this,n,$Heap); 

      /* the following three invariants are essential */
      invariant SubSet(r1, $Heap[this, Subject.O]);
      invariant (forall p : ref :: IsGoodObject(p, $Heap) && r1[p] 
                       ==> Obs(p, this, n, $Heap)); /* notified observers are uptodate */
      invariant EqualSet(UnionSet(List2Observers(node, $Heap), r1), $Heap[this, Subject.O]);

      /* translation needs to generate the following inv. automatically */

      invariant (forall p: ref, f: name ::
                   { $Heap[p,f] }
                   IsGoodObject(p,old($Heap)) ==> 
                         old($Heap)[p,f] == $Heap[p,f] || 
                        (p == this && f == Subject.val) ||
                        (old($Heap)[this,Subject.O][p] && f == Observer.caches)
                ); 
    { 
      call Observer.notify($Heap[node,Node.info], this);
      r1 := UnionSet(r1, SingletonSet($Heap[node,Node.info]));
      node := $Heap[node,Node.nxt];
    }
  }

procedure Subject.get(this: ref) returns ($result : int)
    free requires IsGoodObject(this, $Heap);
    // user-declared preconditions
    requires Sub(this,$Heap[this,Subject.val],$Heap);
    // user-declared postconditions
    ensures Sub(this,$Heap[this,Subject.val],$Heap);
    ensures $result == $Heap[this,Subject.val];
    free ensures NoDeallocs(old($Heap), $Heap);
  {
    $result := $Heap[this,Subject.val];
  }

procedure Subject.add(this: ref where IsGoodObject(this, $Heap), 
                      o: ref where IsGoodObject(o, $Heap))
    // user-declared preconditions
    requires !$Heap[this, Subject.O][o];
    requires Sub(this,$Heap[this,Subject.val],$Heap);

    modifies $Heap;

    // user-declared postconditions
    ensures Sub(this, $Heap[this,Subject.val], $Heap);
    ensures $Heap[this, Subject.O][o];

    ensures EqualSet(UnionSet(SingletonSet(o),old($Heap)[this,Subject.O]), 
                     $Heap[this,Subject.O]) ;
  
    // wr <this>\Dot O, <this>\Dot obs, wr all 
    ensures 
      (forall p: ref, f: name ::
        IsGoodObject(p,old($Heap)) ==>
        old($Heap[p,f]) == $Heap[p,f] ||
          (p == this && f == Subject.O) ||
          (p == this && f == Subject.obs)
      );
    free ensures NoDeallocs(old($Heap), $Heap);
  { 

    var n : ref; 
    $Heap[this,Subject.O] := UnionSet(SingletonSet(o),$Heap[this,Subject.O]);
    assume IsGoodHeap($Heap);

    // Create a new Node
    // n := new Node(o)
      havoc n;
      assume !$Heap[n,alloc]; 
      call Node..ctor(n,o); // after this call IsGoodObject(n) holds

    // Stick it on the front of the list of observers
    $Heap[n,Node.nxt] := $Heap[this,Subject.obs];
    assume IsGoodHeap($Heap);
    $Heap[this,Subject.obs] := n;
    assume IsGoodHeap($Heap);

  }
// END OF SUBJECT

// OBSERVER
procedure Observer..ctor(this: ref where IsGoodObject(s, $Heap), 
                         s : ref where IsGoodObject(this, $Heap))
  // can we relax the following?
  requires (forall p : ref :: IsGoodObject(p, $Heap) ==> !$Heap[p, Subject.O][this]);
  requires Sub(s, $Heap[s,Subject.val], $Heap);
  modifies $Heap;
  // user-declared postconditions
  ensures Sub(s, $Heap[s,Subject.val], $Heap);
  ensures Obs(this, s, $Heap[s,Subject.val], $Heap);

  ensures $Heap[s, Subject.O][this];
  ensures EqualSet(UnionSet(SingletonSet(this), old($Heap)[s, Subject.O]), 
                   $Heap[s, Subject.O]) ; 
  // target object is allocated upon return
  free ensures $Heap[this, alloc]; /* needed for callers of Observer..ctor */
  // <s>\Dot O, wr all // "wr all" is because Subject.register is called
  ensures 
    (forall p: ref, f: name ::
       IsGoodObject(p,old($Heap)) ==>
       old($Heap[p,f]) == $Heap[p,f] || p == this ||
         (p == s && f == Subject.O) ||
         (p == s && f == Subject.obs)          /* needed because of add */
    );
  free ensures NoDeallocs(old($Heap), $Heap);
  {
    call Subject.register(s, this);
  }

procedure Observer.notify(this: ref where IsGoodObject(this, $Heap), 
                          s: ref where IsGoodObject(s, $Heap))
    // user-declared preconditions
    requires $Heap[s, Subject.O][this];
    requires Sub(s, $Heap[s, Subject.val], $Heap);
    modifies $Heap;

    // user-declared postconditions
    ensures Sub(s, $Heap[s, Subject.val], $Heap);
    ensures Obs(this, s, $Heap[s, Subject.val], $Heap);
    // wr <this>.\Dot caches[s]
    ensures 
      (forall p: ref, f: name :: { $Heap[p,f] }
        IsGoodObject(p,old($Heap)) ==>
        old($Heap[p,f]) == $Heap[p,f] || 
           (p == this && f == Observer.caches) 
      );

    ensures 
      (forall otherSubject: ref :: 
         otherSubject != s ==> $Heap[this, Observer.caches][otherSubject] == 
                               old($Heap)[this, Observer.caches][otherSubject]
      );

    free ensures NoDeallocs(old($Heap), $Heap);

  {
    var tmp : int;
    var localCaches : [ref]int;
    call tmp := Subject.get(s);
    localCaches := $Heap[this,Observer.caches];
    localCaches[s] := tmp;
    $Heap[this, Observer.caches] := localCaches;
    assume IsGoodHeap($Heap); // needed after each update to $Heap    
  }

procedure Observer.val(this: ref where IsGoodObject(this, $Heap), 
                       s: ref where IsGoodObject(s, $Heap)) returns ($result : int)
    // user-declared preconditions
    requires SubObs(s, $Heap[s, Subject.val], $Heap);
    requires $Heap[s, Subject.O][this];
    // user-declared postconditions
    ensures SubObs(s, $Heap[s, Subject.val], $Heap);
    ensures $result == $Heap[s, Subject.val];
    free ensures NoDeallocs(old($Heap), $Heap);
  {
    $result := $Heap[this,Observer.caches][s];
  }
// END OF OBSERVER
