// Definitions
// every object has an 'alloc' field, which says whether or not
// the object has been allocated
const unique alloc: <bool>name;

function IsGoodHeap(h : [ref,<x>name]x) returns (bool);

// After this point, some helpful functions

function IsGoodObject(r: ref, h: [ref,<x>name]x) returns (bool);
axiom (forall o: ref, h: [ref,<x>name]x ::
          { IsGoodObject(o,h) }
          IsGoodObject(o,h) <==> IsGoodHeap(h) && o != null && h[o,alloc]);

function IsGoodRegion(r: [ref]bool, h: [ref,<x>name]x) returns (bool);

axiom (forall r: [ref]bool, h: [ref,<x>name]x ::
      { IsGoodRegion(r, h) }
      IsGoodRegion(r, h) <==> 
         (forall o: ref :: r[o] ==> IsGoodObject(o, h))
      );
/* The following axiom is helpful */
axiom (forall r: [ref]bool, o: ref, h: [ref,<x>name]x ::
      { IsGoodRegion(UnionSet(SingletonSet(o), r), h) }
      IsGoodRegion(UnionSet(SingletonSet(o), r), h) <==> 
        IsGoodObject(o, h) && IsGoodRegion(r, h)
      );

function RegionClosure(r: [ref]bool, h: [ref, <x>name]x, field: <ref>name) 
         returns (bool);
// make sure that r is a good region in h
axiom (forall r: [ref]bool, h: [ref, <x>name]x, f: <ref>name :: 
      { RegionClosure(r, h, f) }
           RegionClosure(r,h,f)
           <==>
           (forall o: ref :: 
              r[o] ==> (h[o,f] == null || r[h[o,f]]))
      );

function NoDeallocs(oldHeap: [ref, <x>name]x, newHeap: [ref, <x>name]x) returns (bool);
axiom (forall oldHeap: [ref, <x>name]x, newHeap: [ref, <x>name]x ::
  { NoDeallocs(oldHeap, newHeap) }
  NoDeallocs(oldHeap, newHeap) <==>
    (forall o: ref :: { newHeap[o,alloc] }
      oldHeap[o,alloc] ==> newHeap[o,alloc]));

function AllNewRegion(oldHeap: [ref, <x>name]x, region: [ref]bool) returns (bool);
axiom (forall oldHeap: [ref, <x>name]x, region: [ref]bool ::
  { AllNewRegion(oldHeap, region) }
  AllNewRegion(oldHeap, region) <==>
    (forall o: ref :: { oldHeap[o,alloc] }{ region[o] }
      o != null && region[o] ==> !oldHeap[o,alloc]));

function DifferenceIsNew(oldHeap: [ref, <x>name]x, 
                         oldRegion: [ref]bool, newRegion: [ref]bool) 
         returns (bool);
axiom (forall oldHeap: [ref, <x>name]x, oldRegion: [ref]bool, newRegion: [ref]bool ::
  { DifferenceIsNew(oldHeap, oldRegion, newRegion) }
  DifferenceIsNew(oldHeap, oldRegion, newRegion) <==>
    (forall o: ref :: { oldHeap[o,alloc] }
      o != null && !oldRegion[o] && newRegion[o] ==> !oldHeap[o,alloc]));

function fresh(oldHeap: [ref, <x>name]x, newHeap: [ref, <x>name]x, r: [ref]bool) 
         returns (bool);
axiom (forall oldHeap: [ref, <x>name]x, newHeap: [ref, <x>name]x, r: [ref]bool ::
           { fresh(oldHeap, newHeap, r) }
         fresh(oldHeap, newHeap, r) <==>
           IsGoodHeap(oldHeap) && IsGoodRegion(r, newHeap) && AllNewRegion(oldHeap, r)
      );

axiom (forall oldHeap: [ref, <x>name]x, newHeap: [ref, <x>name]x, 
              r1: [ref]bool, r2: [ref]bool ::
           { fresh(oldHeap, newHeap, UnionSet(r1, r2)) }
         fresh(oldHeap, newHeap, UnionSet(r1, r2)) <==>
           IsGoodHeap(oldHeap) && IsGoodRegion(r1, newHeap) && 
           IsGoodRegion(r2, newHeap) && AllNewRegion(oldHeap, r1) && 
           AllNewRegion(oldHeap, r2)
      );




