//
// Private fields, axioms for functions, and implementations of procedures
//

const unique Stack.contents: <ref>name; // ArrayList<T>

implementation Stack..ctor(this: ref)
{
  var a: ref;
  // a = new ArrayList();
    havoc a;
    assume !Heap[a,alloc];
    call ArrayList..ctor(a);

  Heap[this, Stack.contents] := a;
    
  // Need to manually update the footprint
  Heap[this, Stack.footprint] := Heap[a, ArrayList.footprint];
  Heap[this, Stack.footprint] := UnionSet(Heap[this, Stack.footprint], 
                                          SingletonSet(a));
  Heap[this, Stack.footprint] := UnionSet(Heap[this, Stack.footprint], 
                                               SingletonSet(this));
  assume IsGoodHeap(Heap); // necessary
}

// function Stack.INV(h: [ref, <x>name]x, o: ref) returns (bool);
// INV depends only on the invariants of things in its footprint
axiom
  (forall H: [ref, <x>name]x, o: ref
    :: { Stack.INV(H,o) }

   IsGoodObject(o, H) && ArrayList.INV(H,H[o,Stack.contents])
     && H[o, Stack.footprint][o]
     && H[o, Stack.footprint][H[o, Stack.contents]]
     && SubSet(H[H[o,Stack.contents],ArrayList.footprint], H[o,Stack.footprint])
     // This next clause allows the stack to be verified 
     // relative to the ArrayList client axioms, 
     // not internal axioms
     && !H[H[o, Stack.contents], ArrayList.footprint][o]
     && H[o, Stack.contents] != o // needed because we don't have the types encoded (see Switch)
     && IsGoodObject(H[o, Stack.contents],H)
     <==>
   Stack.INV(H,o)
  );

// function #Stack.size(h: [ref, <x>name]x, o: ref) returns (int);
axiom
  (forall h: [ref, <x>name]x, o: ref
    :: { #Stack.size(h,o) }
   IsGoodObject(o, h)
     ==>
   #Stack.size(h,o) == #ArrayList.size(h,h[o,Stack.contents])
  );

implementation Stack.Push(this: ref, o : int)
{
  call ArrayList.Add(Heap[this,Stack.contents],o);
  // Need to maintain the relationship as part of stack's invariant and since
  // footprint is a field, need to manually update it to keep it a superset of
  // the arraylist's footprint since ArrayList.Add is allowed to grow its footprint.
  Heap[this, Stack.footprint] := UnionSet(Heap[this,Stack.footprint], 
                                          Heap[Heap[this,Stack.contents],
                                               ArrayList.footprint]);
  assume IsGoodHeap(Heap); // necessary
}  

// implementation Stack.Pop(this: ref) returns (result: int)

// implementation Stack.Empty(this: ref) returns (result: bool)

implementation Stack.Switch(this: ref, other: ref)
{
  var tmp: ref;
  var tmpFootprint: [ref]bool;

  tmp := Heap[this,Stack.contents];
  Heap[this,Stack.contents] := Heap[other,Stack.contents];
  Heap[other,Stack.contents] := tmp;
  
  // Since the footprint is a field, need to manually switch the footprints
  Heap[this,Stack.footprint] := UnionSet(Heap[Heap[this,Stack.contents],
                                              ArrayList.footprint], 
                                         SingletonSet(this));
  Heap[this,Stack.footprint] := UnionSet(Heap[this,Stack.footprint], 
                                         SingletonSet(Heap[this,Stack.contents]));
  Heap[other,Stack.footprint] := UnionSet(Heap[Heap[other,Stack.contents], 
                                               ArrayList.footprint], 
                                          SingletonSet(other));
  Heap[other,Stack.footprint] := UnionSet(Heap[other,Stack.footprint], 
                                          SingletonSet(Heap[other,Stack.contents]));

  assume IsGoodHeap(Heap); // necessary

}
