procedure Stack.Client() returns ()
  modifies Heap;
{
    var s1: ref;
    var s2: ref;
    var i : int;

  // s1 = new Stack();
    havoc s1;
    assume !Heap[s1,alloc];
    call Stack..ctor(s1);
assert #Stack.size(Heap,s1) == 0;

  // s2 = new Stack();
    havoc s2;
    assume !Heap[s2,alloc];
    call Stack..ctor(s2);
assert #Stack.size(Heap,s2) == 0;
assert #Stack.size(Heap,s1) == 0;

  call Stack.Push(s1,5);
assert #Stack.size(Heap,s2) == 0;
assert #Stack.size(Heap,s1) == 1;

  call i := Stack.Pop(s1);
assert #Stack.size(Heap,s1) == 0;
//  assert i == 5; // can't assert this because Stack's specs aren't functionally complete
}

procedure Stack.Client2() returns ()
  modifies Heap;
{
    var s1: ref;
    var s2: ref;
    var i : int;

  // s1 = new Stack();
    havoc s1;
    assume !Heap[s1,alloc];
    call Stack..ctor(s1);

  // s2 = new Stack();
    havoc s2;
    assume !Heap[s2,alloc];
    call Stack..ctor(s2);

  call Stack.Push(s1,5);
  
  // meets precondition: DisjointSet(Heap[s,Stack.footprint], Heap[s2,Stack.footprint]);
  call Stack.Switch(s1,s2);
}

procedure Stack.Client3(s: ref) returns ()
  free requires IsGoodObject(s, Heap);
  requires Stack.INV(Heap,s);
  modifies Heap;
  ensures Stack.INV(Heap,s);
{
  var s2: ref;
  var i : int;

  // s2 = new Stack();
    havoc s2;
    assume !Heap[s2,alloc];
    call Stack..ctor(s2);

  call Stack.Push(s,5);
  
  // meets precondition: DisjointSet(Heap[s,Stack.footprint], Heap[s2,Stack.footprint]);
  call Stack.Switch(s,s2);
}

// Expected errors because taking two arbitrary heaps means that one might be within
// the other's footprint. This is because we do not differentiate between the data
// structures used to hold the contents of the stack and the contents themselves
// (i.e., the footprint includes both the array or linked list and the elements pointed
// to from the array or linked list).
// StackClient.bpl(87,3): Error BP5001: This assertion might not hold.
// Execution trace:
//     StackClient.bpl(85,3): anon0
// StackClient.bpl(88,3): Error BP5001: This assertion might not hold.
// Execution trace:
//     StackClient.bpl(85,3): anon0
// StackClient.bpl(90,1): Error BP5003: A postcondition might not hold at this return statement.
// StackClient.bpl(81,3): Related location: This is the postcondition that might not hold.
// 
procedure Stack.Client4(s1: ref, s2: ref) returns ()
  free requires IsGoodObject(s1, Heap);
  free requires IsGoodObject(s2, Heap);
  // user defined preconditions
  requires Stack.INV(Heap,s1);
  requires Stack.INV(Heap,s2);
  modifies Heap;
  ensures Stack.INV(Heap,s1);
  ensures Stack.INV(Heap,s2);
{
  var i : int;

  call Stack.Push(s1,5);
  
  assert !Heap[s2,Stack.footprint][s1];
  assert DisjointSet(Heap[s1,Stack.footprint], Heap[s2,Stack.footprint]);

}
