// Stack<T>

const unique Stack.footprint: <[ref]bool>name;

procedure Stack..ctor(this: ref);
  free requires IsGoodObject(this,Heap);
  modifies Heap;
  // user-declared postconditions
  ensures #Stack.size(Heap,this) == 0;
  // fresh(this.footprint - { this });
  ensures (forall p: ref:: Heap[this,Stack.footprint][p] 
             ==> (p == this || !old(Heap)[p,alloc]));

  // target object is allocated upon return
  free ensures IsGoodObject(this, Heap);
  ensures 
      (forall p: ref, f: name ::
        old(Heap)[p, alloc] ==> old(Heap[p,f]) == Heap[p,f] || p == this 
      );
  ensures Stack.INV(Heap,this);
  free ensures NoDeallocs(old(Heap),Heap);


function Stack.INV(h: [ref, <x>name]x, o: ref) returns (bool);

function #Stack.size(h: [ref, <x>name]x, o: ref) returns (int);

procedure Stack.Push(this: ref, o : int);
  free requires IsGoodObject(this,Heap);
  requires Stack.INV(Heap,this);
  modifies Heap;
  ensures Stack.INV(Heap,this);
  ensures #Stack.size(Heap,this) == #Stack.size(old(Heap),this) + 1;
  // frame condition from "modifies footprint"
  ensures 
      (forall p: ref, f: name ::
        old(Heap)[p, alloc] ==> 
         old(Heap[p,f]) == Heap[p,f] // everything remains unchanged
         || old(Heap)[this,Stack.footprint][p] // or it is an object 
         // contained in self's footprint
      );
  ensures DifferenceIsNew(old(Heap), 
                          old(Heap)[this, Stack.footprint], 
                          Heap[this, Stack.footprint]);
  free ensures NoDeallocs(old(Heap),Heap);

procedure Stack.Pop(this: ref) returns (result: int);
  requires Stack.INV(Heap,this);
  modifies Heap;
  ensures Stack.INV(Heap,this);
  ensures #Stack.size(Heap,this) == #Stack.size(old(Heap),this) - 1;
//  ensures result == old(contents.get(contents.size()-1));

procedure Stack.Empty(this: ref) returns (result: bool);
  requires Stack.INV(Heap,this);
  ensures result == (#Stack.size(Heap,this) == 0);

procedure Stack.Switch(this: ref, other: ref);
  free requires IsGoodObject(this, Heap);
  free requires IsGoodObject(other, Heap);
  
  // user-declared preconditions
  requires Stack.INV(Heap, other);
  requires Stack.INV(Heap,this);
  
  // Note that this *cannot* be replaced by "this != other". Just knowing that
  // the two stacks are different is not equivalent to knowing that somehow
  // one of them isn't stuck inside the other's footprint.
  requires DisjointSet(Heap[this,Stack.footprint],Heap[other,Stack.footprint]);
  
  modifies Heap;
  
  ensures Stack.INV(Heap,this);
  ensures Stack.INV(Heap,other);
  ensures #Stack.size(Heap,this) == #Stack.size(old(Heap), other);
  ensures #Stack.size(Heap,other) == #Stack.size(old(Heap), this);
  ensures DisjointSet(Heap[this,Stack.footprint],Heap[other,Stack.footprint]);
  ensures DifferenceIsNew(old(Heap),
                          UnionSet(old(Heap)[this,Stack.footprint], 
                                   old(Heap)[other,Stack.footprint]),
                          UnionSet(Heap[this,Stack.footprint],
                                   Heap[other,Stack.footprint]));
  // frame condition from "modifies footprint union other.footprint"
  ensures
      (forall p: ref, f: name ::
        old(Heap)[p, alloc] ==> 
         old(Heap[p,f]) == Heap[p,f] // everything remains unchanged
         || old(Heap)[this,Stack.footprint][p] // or it is an object contained 
            //in self's footprint
         || old(Heap)[other,Stack.footprint][p] // or it is 
            //in other's footprint
      );
  free ensures NoDeallocs(old(Heap),Heap);
