// Hiding: This conforms to Section 6.1 of the paper
// Using SubX(s,v,h), SubObsX(s,v,h) and the hidden invariant SubObsH(s,v,h)
var $Heap: [ref, <x>name]x where IsGoodHeap($Heap);

// These are the fields of the class Subject
const unique Subject.obs: <ref>name;
const unique Subject.val: <int>name;
const unique Subject.O: <[ref]bool>name;

// These are the fields of the class Observer
const unique Observer.sub: <ref>name;
const unique Observer.cache: <int>name;
const unique Observer.nxt: <ref>name;

// In Spec#, these would be from the translation of [Pure] methods
// written as part of the program
function List(o : ref, h: [ref,<x>name]x) returns ([ref]bool);
axiom (forall o : ref, h: [ref,<x>name]x :: { List(o,h) }
        (o == null <==> EqualSet(List(o,h),EmptySet))
      );

axiom (forall o : ref, h: [ref,<x>name]x, r: [ref]bool :: { EqualSet(List(o,h), r) } 
        (o != null ==> EqualSet(List(o,h), 
                                UnionSet(SingletonSet(o),List(h[o,Observer.nxt],h))))
      );

axiom (forall H: [ref,<x>name]x, K: [ref,<x>name]x, o : ref ::
         { List(o,H), List(o,K) }
         (forall o2 : ref  ::
             o2 != null && List(o,H)[o2] && List(o,K)[o2]
             ==> H[o2,Observer.nxt] == K[o2,Observer.nxt])
          ==>
          EqualSet(List(o,H), List(o,K))
      );

function SubX(s : ref, v : int, h: [ref,<x>name]x) returns (bool);
axiom (forall s : ref, v : int, h: [ref,<x>name]x :: { SubX(s,v,h) }
        SubX(s,v,h)  <==> (IsGoodObject(s,h) && h[s,Subject.val] == v)
      );

function SubH(s : ref, h: [ref,<x>name]x) returns (bool);
axiom (forall s : ref, h: [ref,<x>name]x :: { SubH(s,h) }
        SubH(s,h)  <==> EqualSet(List(h[s,Subject.obs],h),h[s,Subject.O])
      );
      
function Obs(o : ref, s : ref, v : int, h: [ref,<x>name]x) returns (bool);
axiom (forall o : ref, s : ref, v : int, h: [ref,<x>name]x :: { Obs(o,s,v,h) }
        Obs(o,s,v,h)  <==> IsGoodObject(o,h) && IsGoodObject(s,h) && 
                           h[o,Observer.cache] == v && h[o,Observer.sub] == s
      );
      
function SubObsX(s : ref, v : int, h: [ref,<x>name]x) returns (bool);
axiom (forall s : ref, v : int, h: [ref,<x>name]x :: { SubObsX(s,v,h) }
        SubObsX(s,v,h) <==> SubX(s,v,h) && 
                            (forall o : ref :: IsGoodObject(o,h) && h[s,Subject.O][o] 
                                               ==> Obs(o,s,v,h))
      );

function SubObsH(s : ref, v : int, h: [ref,<x>name]x) returns (bool);
axiom (forall s : ref, v : int, h: [ref,<x>name]x :: { SubObsH(s,v,h) }
        SubObsH(s,v,h) <==> SubX(s,v,h) && SubH(s,h)
      );

axiom (forall h: [ref,<x>name]x, s: ref :: { IsGoodRegion(h[s, Subject.O], h) }
               IsGoodObject(s, h) ==> IsGoodRegion(h[s, Subject.O], h)
      );

axiom (forall h: [ref,<x>name]x, s: ref, o: ref :: { h[s, Subject.O][o] }
               IsGoodObject(s, h) ==> IsGoodRegion(h[s, Subject.O], h)
      );

procedure Subject..ctor(this: ref)

    free requires IsGoodObject(this,$Heap);
    modifies $Heap;
    // user-declared postconditions
    ensures SubObsX(this,0,$Heap);
    ensures EqualSet($Heap[this, Subject.O], EmptySet);
    // target object is allocated upon return
    free ensures $Heap[this, alloc];
    // wr <this>\Dot val, <this>\Dot obs, <this>\Dot O
    ensures 
      (forall p: ref, f: name ::
         IsGoodObject(p,old($Heap)) ==>
         old($Heap[p,f]) == $Heap[p,f] || p == this 
      );
    free ensures NoDeallocs(old($Heap), $Heap);

  {
    $Heap[this,Subject.obs] := null;
    assume IsGoodHeap($Heap);  
    $Heap[this,Subject.val] := 0;
    assume IsGoodHeap($Heap);
    $Heap[this,Subject.O] := EmptySet;
    assume IsGoodHeap($Heap);
  }

procedure Subject.register(this: ref, o: ref)
    free requires IsGoodObject(this,$Heap);
    free requires IsGoodObject(o,$Heap); // implies o != null
    // user-declared preconditions
    requires !$Heap[this, Subject.O][o]; /* to prevent o from being added to the list twice */
    requires SubX(this,$Heap[this,Subject.val],$Heap);
    // not enough to prevent o from already being registered with another subject
    requires $Heap[o,Observer.nxt] == null; 
    //about to register o in above, so require its sub field to point to this
    requires $Heap[o,Observer.sub] == this; 
    requires SubObsH(this, $Heap[this,Subject.val],$Heap);

    modifies $Heap;
    // user-declared postconditions

    ensures SubX(this,$Heap[this,Subject.val],$Heap);
    ensures Obs(o,this,$Heap[this,Subject.val],$Heap);
    ensures EqualSet(UnionSet(SingletonSet(o),old($Heap)[this,Subject.O]), 
                              $Heap[this,Subject.O]) ;

    ensures SubObsH(this, $Heap[this,Subject.val],$Heap);
    // wr this.O\Dot nxt, <this>\Dot O, <o>\Dot nxt, <o>\Dot cache, <this>\Dot obs
    ensures 
      (forall p: ref, f: name ::
        IsGoodObject(p,old($Heap)) ==>
        old($Heap[p,f]) == $Heap[p,f] || 
          (old($Heap)[this,Subject.O][p] && f == Observer.nxt) ||
          (p == this && f == Subject.O) ||
          (p == o && f == Observer.nxt) ||
          (p == o && f == Observer.cache) ||
          (p == this && f == Subject.obs)
      );
    free ensures NoDeallocs(old($Heap), $Heap);
  {
    call Subject.add(this,o);
    call Observer.notify(o);

  }

procedure Subject.update(this: ref, n: int)
    // user-declared preconditions
    free requires IsGoodObject(this, $Heap);
    requires SubObsX(this,$Heap[this,Subject.val],$Heap);
    requires SubObsH(this,$Heap[this,Subject.val],$Heap);

    modifies $Heap;
    // user-declared postconditions
    ensures SubObsX(this, n, $Heap);
    ensures SubObsH(this, $Heap[this,Subject.val],$Heap);

    ensures
      (forall p: ref, f: name ::
        IsGoodObject(p,old($Heap)) ==>
        old($Heap[p,f]) == $Heap[p,f] || 
          (p == this && f == Subject.val) ||
          (old($Heap)[this,Subject.O][p] && f == Observer.cache)
      );
    free ensures NoDeallocs(old($Heap), $Heap);
  {
    var o : ref where (o == null) || IsGoodObject(o, $Heap); 
    var r1 : [ref]bool where IsGoodRegion(r1, $Heap);

    $Heap[this,Subject.val] := n;
    assume IsGoodHeap($Heap);
    o := $Heap[this,Subject.obs];
    r1 := EmptySet;

    while (o != null)
      invariant SubX(this,n,$Heap);
      /* the following invariants are essential */
      invariant SubSet(r1, $Heap[this, Subject.O]);
      invariant (forall p : ref :: IsGoodObject(p, $Heap) && r1[p] 
                       ==> Obs(p, this, n, $Heap)); /* notified observers are uptodate */
      invariant EqualSet(UnionSet(List(o, $Heap), r1), $Heap[this, Subject.O]);
      invariant SubObsH(this, $Heap[this,Subject.val],$Heap);

      /* translation needs to generate the following inv. automatically */
      invariant (forall p: ref, f: name ::
                   { $Heap[p,f] } // this trigger is unnecessary
                   IsGoodObject(p,old($Heap)) ==> 
                         old($Heap[p,f]) == $Heap[p,f] || 
                        (p == this && f == Subject.val) ||
                        (old($Heap)[this,Subject.O][p] && f == Observer.cache)); 

    { 
      call Observer.notify(o);
      r1 := UnionSet(r1, SingletonSet(o));
      o := $Heap[o,Observer.nxt];
    }
  }

procedure Subject.get(this: ref) returns ($result : int)
    free requires IsGoodObject(this, $Heap);
    // user-declared preconditions
    requires SubX(this,$Heap[this,Subject.val],$Heap);
    requires SubObsH(this, $Heap[this,Subject.val],$Heap);
    // user-declared postconditions
    ensures SubX(this,$Heap[this,Subject.val],$Heap);
    ensures $result == $Heap[this,Subject.val];
    ensures SubObsH(this, $Heap[this,Subject.val],$Heap);
    
  {
    $result := $Heap[this,Subject.val];
  }

procedure Subject.add(this: ref, o: ref)
    // user-declared preconditions
    requires IsGoodObject(o, $Heap); // implies o != null;
    requires !$Heap[this, Subject.O][o];
    requires SubX(this,$Heap[this,Subject.val],$Heap);
    requires $Heap[o,Observer.nxt] == null;
    requires $Heap[o, Observer.sub] == this; //this is critical
    requires SubObsH(this, $Heap[this,Subject.val],$Heap);
    modifies $Heap;

    // user-declared postconditions
    ensures SubX(this,$Heap[this,Subject.val],$Heap);
    ensures EqualSet(UnionSet(SingletonSet(o),old($Heap)[this,Subject.O]), 
                     $Heap[this,Subject.O]) ;
    ensures SubObsH(this, $Heap[this,Subject.val],$Heap);

    // wr this.O\Dot nxt, <this>\Dot O, <o>\Dot nxt, <this>\Dot obs
    ensures 
      (forall p: ref, f: name ::
        IsGoodObject(p,old($Heap)) ==>
        old($Heap)[p,f] == $Heap[p,f] ||
          (old($Heap)[this,Subject.O][p] && f == Observer.nxt) ||
          (p == o && f == Observer.nxt) ||
          (p == this && f == Subject.O) ||
          (p == this && f == Subject.obs)
      );
    free ensures NoDeallocs(old($Heap), $Heap);
  { 
    $Heap[this,Subject.O] := UnionSet(SingletonSet(o),$Heap[this,Subject.O]);
    assume IsGoodHeap($Heap);
    // Add o to the front of the list of observers
    $Heap[o,Observer.nxt] := $Heap[this,Subject.obs];
    assume IsGoodHeap($Heap);
    $Heap[this,Subject.obs] := o;
    assume IsGoodHeap($Heap);
  }

procedure Observer..ctor(this: ref, s : ref)
    free requires IsGoodObject(s, $Heap);
    free requires IsGoodObject(this,$Heap); 
    // above needed because we are passing self to register */
    requires (forall p : ref :: IsGoodObject(p, $Heap) ==> !$Heap[p, Subject.O][this]);
    requires SubX(s, $Heap[s,Subject.val], $Heap);
    requires SubObsH(s, $Heap[s,Subject.val],$Heap);
    modifies $Heap;
    // user-declared postconditions
    ensures SubX(s, $Heap[s,Subject.val], $Heap);
    ensures Obs(this, s, $Heap[s,Subject.val], $Heap);

    ensures $Heap[s,Subject.O][this];
    ensures EqualSet(UnionSet(SingletonSet(this), old($Heap)[s,Subject.O]), 
                     $Heap[s,Subject.O]) ; //needed for client

    ensures SubObsH(s, $Heap[s,Subject.val],$Heap);
    // wr s.O\Dot nxt, <s>\Dot O
    // Since the only nxt field that is changed is "this" one, don't need to own
    // up to it here in the modifies clause.
    ensures 
      (forall p: ref, f: name ::
        IsGoodObject(p,old($Heap)) ==>
        old($Heap)[p,f] == $Heap[p,f] || p == this ||
          (old($Heap)[s,Subject.O][p] && f == Observer.nxt) ||
          (p == s && f == Subject.O) ||
          (p == s && f == Subject.obs)          /* needed because of add */
      );
    free ensures NoDeallocs(old($Heap), $Heap);
    {
      // observer takes responsibility of setting its sub field 
      $Heap[this,Observer.sub] := s; 
      assume IsGoodHeap($Heap);
      $Heap[this,Observer.nxt] := null;
      assume IsGoodHeap($Heap); 
      call Subject.register(s,this); // pre for this call is this.sub == s
    }

procedure Observer.notify(this: ref)
    free requires IsGoodObject(this,$Heap);
    // user-declared preconditions
    requires SubX($Heap[this,Observer.sub],
                  $Heap[$Heap[this,Observer.sub],Subject.val],$Heap);
    requires SubObsH($Heap[this,Observer.sub],
                     $Heap[$Heap[this,Observer.sub],Subject.val],$Heap);
    modifies $Heap;
    // user-declared postconditions
    ensures SubX($Heap[this,Observer.sub],
                 $Heap[$Heap[this,Observer.sub],Subject.val],$Heap);
    ensures Obs(this,$Heap[this,Observer.sub],
                $Heap[$Heap[this,Observer.sub],Subject.val], $Heap);
    // wr <this>.\Dot cache
    ensures 
      (forall p: ref, f: name :: { $Heap[p,f] }
        IsGoodObject(p,old($Heap)) ==>
        old($Heap[p,f]) == $Heap[p,f] || 
           (p == this && f == Observer.cache) 
      );
    free ensures NoDeallocs(old($Heap), $Heap);
    {
      var tmp : int;
      call tmp := Subject.get($Heap[this,Observer.sub]);
      $Heap[this,Observer.cache] := tmp;
      assume IsGoodHeap($Heap); // needed after each update to $Heap
    }

procedure Observer.val(this: ref) returns ($result : int)
    free requires IsGoodObject(this,$Heap);
    // user-declared preconditions
    requires SubObsX($Heap[this,Observer.sub], 
                     $Heap[$Heap[this,Observer.sub],Subject.val],$Heap);
    requires $Heap[$Heap[this,Observer.sub],Subject.O][this];
    // user-declared postconditions
    ensures SubObsX($Heap[this,Observer.sub], 
                    $Heap[$Heap[this,Observer.sub],Subject.val],$Heap);
    ensures $result == $Heap[$Heap[this,Observer.sub],Subject.val];
    {
      $result := $Heap[this,Observer.cache];
    }
